﻿using System;
using System.Collections.Generic;

// Also see:
// https://github.com/dotnet/corefx/blob/master/src/System.Linq/src/System/Linq/Enumerable.cs

namespace AdrianKousz.Util
{
	public static class Enumerables
	{
		#region Fillers

		public static T[] Fill<T>(this T[] array) where T : new() { return Fill(array, () => new T()); }
		public static T[] Fill<T>(this T[] array, T value) { return Fill(array, () => value); }
		public static T[] Fill<T>(this T[] array, Func<T> f)
		{
			var len = array.Length;
			for (var i = 0; i < len; i++) array[i] = f();
			return array;
		}

		public static ICollection<T> Fill<T>(this ICollection<T> list, int count) where T : new() { return Fill(list, count, () => new T()); }
		public static TColl Fill<T, TColl>(this TColl list, int count, T value) where TColl : ICollection<T> { return Fill(list, count, () => value); }
		public static TColl Fill<T, TColl>(this TColl list, int count, Func<T> f) where TColl : ICollection<T>
		{
			for (var i = 0; i < count; i++) list.Add(f());
			return list;
		}

		#endregion

		#region Type Helpers

		public static T[] CreateArray<T>(this T example, int count)
		{
			return new T[count];
		}

		public static List<T> CreateList<T>(this T example, int count = 12)
		{
			return new List<T>(count);
		}

		public static T[] NewArray<T>(this IEnumerable<T> source, int count)
		{
			return new T[count];
		}

		public static List<T> NewList<T>(this IEnumerable<T> source, int count = 12)
		{
			return new List<T>(count);
		}

		public static List<T> ToList<T>(this IEnumerable<T> source)
		{
			return source as List<T> ?? new List<T>(source);
		}

		public static T[] ToArray<T>(this IEnumerable<T> source)
		{
			return source as T[] ?? ToList(source).ToArray();
		}

		#endregion

		#region Functional Classics

		public static int IndexOf<T>(this IEnumerable<T> source, T value, Func<T, T, bool> comparer)
		{
			var i = 0;
			foreach (var el in source) {
				if (comparer(el, value)) return i;
				i++;
			}
			return -1;
		}

		public static void ForEach<T>(this IEnumerable<T> source, Action<T> f)
		{
			foreach (var el in source) f(el);
		}

		public static void ForEach<K, V>(this IDictionary<K, V> source, Action<K, V> f)
		{
			foreach (var el in source) f(el.Key, el.Value);
		}

		public static void ForBoth<T1, T2>(this IEnumerable<T1> s1, IEnumerable<T2> s2, Action<T1, T2> f)
		{
			using (var en1 = s1.GetEnumerator())
			using (var en2 = s2.GetEnumerator())
			{
				while (en1.MoveNext() && en2.MoveNext())
					f(en1.Current, en2.Current);
			}
		}

		public static IEnumerable<TResult> Map<T, TResult>(this IEnumerable<T> source, Func<T, TResult> f)
		{
			var result = GetList<T, TResult>(source);
			foreach (var el in source) result.Add(f(el));
			return result;
		}

		public static IEnumerable<T> Filter<T>(this IEnumerable<T> source, Func<T, bool> f)
		{
			var result = GetList<T, T>(source);
			foreach (var el in source) if (f(el)) result.Add(el);
			return result;
		}

		public static TResult FoldL<T, TResult>(this IEnumerable<T> source, TResult initial, Action<TResult, T> f)
		{ return FoldL(source, initial, (folded, el) => { f(folded, el); return folded; }); }
		public static TResult FoldL<T, TResult>(this IEnumerable<T> source, TResult initial, Func<TResult, T, TResult> f)
		{
			var result = initial;
			foreach (var el in source) result = f(result, el);
			return result;
		}

		public static T FoldL1<T>(this IEnumerable<T> source, Func<T, T, T> f)
		{
			using (var en = source.GetEnumerator())
			{
				if (!en.MoveNext()) throw new InvalidOperationException("No elements");
				var result = en.Current;
				while (en.MoveNext()) result = f(result, en.Current);
				return result;
			}
		}

		public static IDictionary<K, V> ToDictionary<K, V>(this IEnumerable<K> ks, IEnumerable<V> vs)
		{
			var result = new Dictionary<K, V>();
			ForBoth(ks, vs, (k, v) => { result[k] = v; });
			return result;
		}

		public static IDictionary<T, T> ToDictionary<T>(this IEnumerable<T> source)
		{
			var result = new Dictionary<T, T>();
			var iskey = true;
			var k = default(T);
			foreach (var el in source)
				if (iskey = !iskey) result[k] = el;
				else k = el;
			return result;
		}

		public static IEnumerable<TResult> Zip<T1, T2, TResult>(this IEnumerable<T1> s1, IEnumerable<T2> s2, Func<T1, T2, TResult> f)
		{
			var result = GetList<T1, TResult>(s1);
			ForBoth(s1, s2, (e1, e2) => { result.Add(f(e1, e2)); });
			return result;
		}

		#endregion

		#region Implementation Detail

		private static List<TResult> GetList<T, TResult>(IEnumerable<T> source)
		{
			var collection = source as ICollection<T>;
			var length = collection == null ? 4 : collection.Count;
			return new List<TResult>(length);
		}

		#endregion
	}
}
